/*
   This structures and functions provide functions
   in DLLs (and shared objects/libraries) with
   access to vibrant and own interface functions
   linked in the kernel.
*/

#define LIB_IINIT_FUNC_NAME     _ContIInit
#define LIB_IINIT_FUNC_STR      "_ContIInit"
#define LIB_IINIT_FUNC_TYPE     _ContIInit_

#define LIB_ITERM_FUNC_NAME     _ContITerm
#define LIB_ITERM_FUNC_STR      "_ContITerm"
#define LIB_ITERM_FUNC_TYPE     _ContITerm_

#if _WIN && !defined(WIN32)
/*
   Types of real callbacks passed from DLLs
   to the umbrella functions provided by the
   kernel for interface vibrant functions.
   These typedefs are copied from vibtypes.h.
   The only difference is that pointers have
   the LIBCALL attribute to allow calls of
   real callbacks defined in DLLs from
   the umbrella callbacks provided by the kernel.
   In MS Windows call to a function in a DLL
   requires the PASCAL method of parameter passing.
*/
#endif

#if _WIN && !defined(WIN32)
/* Vibrant */
typedef  void  (LIBCALL *iNlm_VoidProc) PROTO((void));
typedef  void  (LIBCALL *iNlm_KeyProc)  PROTO((Nlm_Char));
typedef  void  (LIBCALL *iNlm_ActnProc) PROTO((Nlm_GraphiC));
typedef  void  (LIBCALL *iNlm_ClckProc) PROTO((Nlm_GraphiC, Nlm_PoinT));
typedef  void  (LIBCALL *iNlm_CharProc) PROTO((Nlm_GraphiC, Nlm_Char));
typedef  void  (LIBCALL *iNlm_ScrlProc) PROTO((Nlm_BaR, Nlm_GraphiC, Nlm_Int2, Nlm_Int2));
typedef  void  (LIBCALL *iNlm_ChngProc) PROTO((Nlm_GraphiC, Nlm_Int2, Nlm_Int2));
typedef  void  (LIBCALL *iNlm_FreeProc) PROTO((Nlm_GraphiC, Nlm_VoidPtr));

typedef  void  (LIBCALL *iNlm_BtnActnProc) PROTO((Nlm_ButtoN));
typedef  void  (LIBCALL *iNlm_ChsActnProc) PROTO((Nlm_ChoicE));
typedef  void  (LIBCALL *iNlm_GrpActnProc) PROTO((Nlm_GrouP));
typedef  void  (LIBCALL *iNlm_IcnActnProc) PROTO((Nlm_IcoN));
typedef  void  (LIBCALL *iNlm_ItmActnProc) PROTO((Nlm_IteM));
typedef  void  (LIBCALL *iNlm_LstActnProc) PROTO((Nlm_LisT));
typedef  void  (LIBCALL *iNlm_MnuActnProc) PROTO((Nlm_MenU));
typedef  void  (LIBCALL *iNlm_PnlActnProc) PROTO((Nlm_PaneL));
typedef  void  (LIBCALL *iNlm_PupActnProc) PROTO((Nlm_PopuP));
typedef  void  (LIBCALL *iNlm_RptActnProc) PROTO((Nlm_RepeaT));
typedef  void  (LIBCALL *iNlm_TxtActnProc) PROTO((Nlm_TexT));
typedef  void  (LIBCALL *iNlm_WndActnProc) PROTO((Nlm_WindoW));
typedef  void  (LIBCALL *iNlm_WndFreeProc) PROTO((Nlm_WindoW, Nlm_VoidPtr));

typedef  void  (LIBCALL *iNlm_PnlClckProc) PROTO((Nlm_PaneL, Nlm_PoinT));
typedef  void  (LIBCALL *iNlm_RptClckProc) PROTO((Nlm_RepeaT, Nlm_PoinT));
typedef  void  (LIBCALL *iNlm_IcnClckProc) PROTO((Nlm_IcoN, Nlm_PoinT));

typedef  void  (LIBCALL *iNlm_BarScrlProc) PROTO((Nlm_BaR, Nlm_GraphiC, Nlm_Int2, Nlm_Int2));
typedef  void  (LIBCALL *iNlm_SltScrlProc) PROTO((Nlm_BaR, Nlm_SlatE, Nlm_Int2, Nlm_Int2));

typedef  void  (LIBCALL *iNlm_IcnChngProc) PROTO((Nlm_IcoN, Nlm_Int2, Nlm_Int2));
typedef  void  (LIBCALL *iNlm_SwtChngProc) PROTO((Nlm_SwitcH, Nlm_Int2, Nlm_Int2));

typedef  void  (LIBCALL *iNlm_SltCharProc) PROTO((Nlm_SlatE, Nlm_Char));

/* DoC */
typedef Nlm_CharPtr (LIBCALL *iNlm_DocPrntProc)  PROTO((Nlm_DoC, Nlm_Int2, Pointer));
typedef void        (LIBCALL *iNlm_DocClckProc)  PROTO((Nlm_DoC, Nlm_PoinT));
typedef void        (LIBCALL *iNlm_DocNotfyProc) PROTO((Nlm_DoC, Nlm_Int2, Nlm_Int2, Nlm_Int2, Nlm_Boolean));
typedef void        (LIBCALL *iNlm_DocDrawProc)  PROTO((Nlm_DoC, Nlm_RectPtr, Nlm_Int2, Nlm_Int2));
typedef Nlm_Boolean (LIBCALL *iNlm_DocShadeProc) PROTO((Nlm_DoC, Nlm_Int2, Nlm_Int2, Nlm_Int2));
typedef void        (LIBCALL *iNlm_DocDataProc)  PROTO((Nlm_DoC, Pointer));
typedef void        (LIBCALL *iNlm_DocPanProc)   PROTO((Nlm_DoC));
typedef void        (LIBCALL *iNlm_DocFreeProc)  PROTO((Nlm_DoC, Nlm_VoidPtr));

typedef void        (LIBCALL *iNlm_DocPutProc)   PROTO((Nlm_DoC, Nlm_Int2, Nlm_CharPtr));
typedef Nlm_CharPtr (LIBCALL *iNlm_DocGetProc)   PROTO((Nlm_DoC, Nlm_Int2));
typedef void        (LIBCALL *iNlm_DocUpdProc)   PROTO((Nlm_DoC, Nlm_Int2, Nlm_Int2));

#else		/* _WIN */

/* Vibrant */
#define iNlm_VoidProc Nlm_VoidProc
#define iNlm_KeyProc  Nlm_KeyProc
#define iNlm_ActnProc Nlm_ActnProc
#define iNlm_ClckProc Nlm_ClckProc
#define iNlm_CharProc Nlm_CharProc
#define iNlm_ScrlProc Nlm_ScrlProc
#define iNlm_ChngProc Nlm_ChngProc
#define iNlm_FreeProc Nlm_FreeProc

#define iNlm_BtnActnProc Nlm_BtnActnProc
#define iNlm_ChsActnProc Nlm_ChsActnProc
#define iNlm_GrpActnProc Nlm_GrpActnProc
#define iNlm_IcnActnProc Nlm_IcnActnProc
#define iNlm_ItmActnProc Nlm_ItmActnProc
#define iNlm_LstActnProc Nlm_LstActnProc
#define iNlm_MnuActnProc Nlm_MnuActnProc
#define iNlm_PnlActnProc Nlm_PnlActnProc
#define iNlm_PupActnProc Nlm_PupActnProc
#define iNlm_RptActnProc Nlm_RptActnProc
#define iNlm_TxtActnProc Nlm_TxtActnProc
#define iNlm_WndActnProc Nlm_WndActnProc
#define iNlm_WndFreeProc Nlm_WndFreeProc

#define iNlm_PnlClckProc Nlm_PnlClckProc
#define iNlm_RptClckProc Nlm_RptClckProc
#define iNlm_IcnClckProc Nlm_IcnClckProc

#define iNlm_BarScrlProc Nlm_BarScrlProc
#define iNlm_SltScrlProc Nlm_SltScrlProc

#define iNlm_IcnChngProc Nlm_IcnChngProc
#define iNlm_SwtChngProc Nlm_SwtChngProc

#define iNlm_SltCharProc Nlm_SltCharProc

/* DoC */
#define iNlm_DocPrntProc Nlm_DocPrntProc
#define iNlm_DocClckProc Nlm_DocClckProc
#define iNlm_DocNotfyProc Nlm_DocNotfyProc
#define iNlm_DocDrawProc Nlm_DocDrawProc
#define iNlm_DocShadeProc Nlm_DocShadeProc
#define iNlm_DocDataProc Nlm_DocDataProc
#define iNlm_DocPanProc Nlm_DocPanProc
#define iNlm_DocFreeProc Nlm_DocFreeProc

#define iNlm_DocPutProc Nlm_DocPutProc
#define iNlm_DocGetProc Nlm_DocGetProc
#define iNlm_DocUpdProc Nlm_DocUpdProc

#endif	/* _WIN */

/*
   The structure passed to the _ContIInit function (if any)
   immediately after loading of any library.
   The structure is used by library's functions to access
   vibrant and own functions linked in the kernel.
*/

typedef struct {
  /* Vibrant */
  RectPtr	screenRect_;
  BoolPtr	dblClick_;
  Int2Ptr	sysCharWidth_;
  Int2Ptr	sysLineHeight_;
  Int2Ptr	sysDescent_;
  Int2Ptr	sysAscent_;
  Int2Ptr	vScrollBarWidth_;
  BoolPtr	Nlm_hasColor_;
  FonT		*ProgramFont_;
  FonT		*SystemFont_;
  Int2Ptr	dialogTextHeight_;
  Int2		(*StringWidth_)(CharPtr ptr);
  Int2		(*TextWidth_)(CharPtr text, size_t len);
  WindoW	(*DocumentWindow_)(Int2 left, Int2 top, Int2 width, Int2 height, CharPtr title, iNlm_WndActnProc close, iNlm_WndActnProc resize);
  WindoW	(*FixedWindow_)(Int2 left, Int2 top, Int2 width, Int2 height, CharPtr title, iNlm_WndActnProc close);
  void		(*SetWindowExtra_)(WindoW w, VoidPtr data, iNlm_WndFreeProc cleanup);
  VoidPtr	(*GetWindowExtra_)(WindoW w);
  WindoW	(*ParentWindow_)(Handle a);
  WindoW	(*CurrentWindow_)(void);
  void		(*UseWindow_)(WindoW w);
  GrouP		(*NormalGroup_)(GrouP prnt, Int2 width, Int2 height, CharPtr title, FonT font, GrpActnProc actn);
  GrouP		(*HiddenGroup_)(GrouP prnt, Int2 width, Int2 height, iNlm_GrpActnProc actn);
  void		(*SetGroupMargins_)(GrouP g, Int2 xMargin, Int2 yMargin);
  void		(*SetGroupSpacing_)(GrouP g, Int2 xSpacing, Int2 ySpacing);
  void		(*Show_)(Handle a);
  void		(*Hide_)(Handle a);
  void		(*Enable_)(Handle a);
  void		(*Disable_)(Handle a);
  Boolean	(*Visible_)(Handle a);
  void		(*Select_)(Handle a);
  void		(*SetValue_)(Handle a, Int2 value);
  Int2		(*GetValue_)(Handle a);
  void		(*SetTitle_)(Handle a, CharPtr title);
  void		(*GetTitle_)(Handle a, CharPtr title, size_t maxsize);
  void		(*SetStatus_)(Handle a, Boolean status);
  Boolean	(*GetStatus_)(Handle a);
  void		(*SetPosition_)(Handle a, RectPtr r);
  void		(*GetPosition_)(Handle a, RectPtr r);
  void		(*SetOffset_)(Handle a, Int2 horiz, Int2 vert);
  void		(*GetOffset_)(Handle a, Int2Ptr horiz, Int2Ptr vert);
  void		(*Reset_)(Handle a);
  Handle	(*Remove_)(Handle a);
  void		(*SetNextPosition_)(Handle a, PoinT nps);
  void		(*GetNextPosition_)(Handle a, PointPtr nps);
  void		(*Advance_)(Handle a);
  void		(*Break_)(Handle a);
  TexT		(*DialogText_)(GrouP prnt, CharPtr dfault, Int2 charWidth, iNlm_TxtActnProc actn);
  TexT		(*HiddenText_)(GrouP prnt, CharPtr dfault, Int2 charWidth, TxtActnProc actn, TxtActnProc tabProc, TxtActnProc rtnProc);
  size_t	(*TextLength_)(TexT t);
  void		(*SelectText_)(TexT t, Int2 begin, Int2 end);
  void		(*TextSelectionRange_)(TexT t, Int2Ptr begin, Int2Ptr end);
  void		(*SetTextSelect_)(TexT t, TxtActnProc slct, TxtActnProc dslct);
  LisT		(*SingleList_)(GrouP prnt, Int2 width, Int2 height, iNlm_LstActnProc actn);
  void		(*ListItem_)(LisT l, CharPtr title);
  PopuP		(*PopupList_)(GrouP prnt, Boolean macLike, PupActnProc actn);
  void		(*PopupItem_)(PopuP p, CharPtr title);
  ButtoN	(*PushButton_)(GrouP prnt, CharPtr title, iNlm_BtnActnProc actn);
  ButtoN	(*DefaultButton_)(GrouP prnt, CharPtr title, iNlm_BtnActnProc actn);
  ButtoN	(*CheckBox_)(GrouP prnt, CharPtr title, iNlm_BtnActnProc actn);
  ButtoN	(*RadioButton_)(GrouP prnt, CharPtr title);
  PrompT	(*StaticPrompt_)(GrouP prnt, CharPtr title, Int2 pixwidth, Int2 pixheight, FonT font, Char just);
  void CDECL	(*AlignObjects_)(int align, ...);
  PaneL		(*SimplePanel_)(GrouP prnt, Int2 pixwidth, Int2 pixheight, iNlm_PnlActnProc draw);
  BaR		(*GetSlateVScrollBar_)(SlatE s);
  void		(*SetPanelClick_)(PaneL p, iNlm_PnlClckProc click, iNlm_PnlClckProc drag, iNlm_PnlClckProc hold, iNlm_PnlClckProc release);
  BaR		(*ScrollBar_)(GrouP prnt, Int2 width, Int2 height, BarScrlProc  actn);
  void		(*CorrectBarValue_)(BaR bar, Int4 val);
  void		(*CorrectBarMax_)(BaR bar, Int4 max);
  void		(*SetRange_)(Handle a, Int2 pgUp, Int2 pgDn, Int2 max);
  void		(*SetObjectExtra_)(Handle a, VoidPtr data, iNlm_FreeProc cleanup);
  VoidPtr	(*GetObjectExtra_)(Handle a);
  Handle	(*Parent_)(Handle a);
  void		(*SetAction_)(Handle a, ActnProc actn);
  void		(*Update_)(void);
  void		(*InvalRect_)(RectPtr r);
  void		(*PaintRect_)(RectPtr r);
  void		(*ClipRect_)(RectPtr r);
  void		(*ResetClip_)(void);
  void		(*OffsetRect_)(RectPtr r, Int2 dx, Int2 dy);
  void		(*LoadRect_)(RectPtr r, Int2 lf, Int2 tp, Int2 rt, Int2 bt);
  void		(*InsetRect_)(RectPtr r, Int2 dx, Int2 dy);
  void		(*UpsetRect_)(RectPtr r, Int2 lf, Int2 tp, Int2 rt, Int2 bt);
  Boolean	(*UnionRect_)(RectPtr src1, RectPtr src2, RectPtr dst);
  Boolean	(*SectRect_)(RectPtr src1, RectPtr src2, RectPtr dst);
  Boolean	(*PtInRect_)(PoinT pt, RectPtr r);
  void		(*LoadPt_)(PointPtr pt, Int2 x, Int2 y);
  void		(*AddPt_)(PoinT src, PointPtr dst);
  void		(*SubPt_)(PoinT src, PointPtr dst);
  Boolean	(*EqualPt_)(PoinT p1, PoinT p2);
  void		(*FrameOval_)(RectPtr r);
  void		(*PaintOval_)(RectPtr r);
  void		(*FrameArc_)(RectPtr r, PoinT start, PoinT end);
  void		(*MoveTo_)(Int2 x, Int2 y);
  void		(*LineTo_)(Int2 x, Int2 y);
  void		(*SetPen_)(PoinT pt);
  void		(*DrawLine_)(PoinT pt1, PoinT pt2);
  void		(*SelectColor_)(Uint1 red, Uint1 green, Uint1 blue);
  Uint4		(*GetColor_)(void);
  void		(*InvertColors_)(void);
  void		(*DecodeColor_)(Uint4 color, Uint1Ptr red, Uint1Ptr green, Uint1Ptr blue);
  void		(*Black_)(void);
  void		(*Red_)(void);
  void		(*Green_)(void);
  void		(*Blue_)(void);
  void		(*Cyan_)(void);
  void		(*Magenta_)(void);
  void		(*Yellow_)(void);
  void		(*White_)(void);
  void		(*WidePen_)(Int2 width);
  void		(*SetPenDash_)(Uint1 offset, Uint1 dash, Uint1 gap);
  void		(*Solid_)(void);
  void		(*Dotted_)(void);
  void		(*Dashed_)(void);
  FonT		(*ParseFont_)(CharPtr spec);
  void		(*SelectFont_)(Nlm_FonT f);
  FonT		(*DeleteFont_)(FonT font);
  Int2		(*Ascent_)(void);
  Int2		(*Descent_)(void);
  Int2		(*Leading_)(void);
  Int2		(*FontHeight_)(void);
  Int2		(*LineHeight_)(void);
  void		(*PaintString_)(CharPtr text);
  void		(*DrawString_)(RectPtr r, CharPtr text, Char jst, Boolean gray);
  void		CDECL (*PaintText_)(char *format, ...);
  void		(*RealizeWindow_)(WindoW w);
  void		(*InvertMode_)(void);
  void		(*CopyMode_)(void);
  void		(*FrameRect_)(RectPtr r);
  void		(*CopyBits_)(RectPtr r, VoidPtr source);
#if MYSWITCH
  Switch	(*SwitchCreate_)(GrouP prnt, Int2 maxval, iNlm_ChngProc actn);
  void		(*SwitchSetValue_)(Switch s, Int2 val);
  Int2		(*SwitchGetValue_)(Switch s);
  void		(*SwitchSetMax_)(Switch s, Int2 maxv);
  void		(*SwitchSetParams_)(Switch s, Int2 val, Int2 maxv);
  Int2		(*SwitchHeight_)(void);
#else
  SwitcH	(*UpDownSwitch_)(GrouP prnt, Boolean text, iNlm_SwtChngProc actn);
  SwitcH	(*LeftRightSwitch_)(GrouP prnt, Boolean text, iNlm_SwtChngProc actn);
  void		(*SetSwitchMax_)(SwitcH s, Int2 max);
  Int2		(*GetSwitchMax_)(SwitcH s);
  void		(*SetSwitchParams_)(SwitcH s, Int2 value, Int2 max);
#endif
  DoC		(*DocumentPanel_)(GrouP prnt, Int2 pixwidth, Int2 pixheight);
  void		(*UpdateDocument_)(DoC doc, Int2 from, Int2 to);
  void		(*AppendText_)(DoC d, CharPtr text, ParPtr parFmt, ColPtr colFmt, FonT font);
  void		(*ReplaceText_)(DoC d, Int2 item, CharPtr text, ParPtr parFmt, ColPtr colFmt, FonT font);
  void		(*DeleteItem_)(DoC d, Int2 item);
  void		(*AdjustDocScroll_)(DoC d);
  void		(*SetDocShade_)(DoC d, DocDrawProc draw, DocShadeProc gray, DocShadeProc invert, DocShadeProc color);
  void		(*GetDocParams_)(DoC d, Int2Ptr numItems, Int2Ptr numLines);
  void		(*GetColParams_)(DoC d, Int2 item, Int2 col, Int2Ptr pixPos, Int2Ptr pixWidth, Int2Ptr pixInset, CharPtr just);
  Boolean	(*ItemIsVisible_)(DoC d, Int2 item, Int2Ptr top, Int2Ptr bottom, Int2Ptr firstLine);
  Boolean	(*GetScrlParams_)(DoC d, Int2Ptr offset, Int2Ptr firstShown, Int2Ptr firstLine);
  void		(*ProcessUpdatesFirst_)(Boolean val);
  Boolean	(*GetInputFileName_)(CharPtr fileName, size_t maxsize, CharPtr extType);
  Boolean	(*GetOutputFileName_)(CharPtr fileName, size_t maxsize, CharPtr dfault);
  Boolean	(*ChooseColorDialog_)(Uint1 PNTR redptr, Uint1 PNTR greenptr, Uint1 PNTR blueptr, Boolean Initial);
  /* Own kernel functions */
  /* dynamic libraries / shared objects / shared librraies, etc. */
  void		(*LibraryUnlockRequest)(LibHandle lib, LibTermProc);
  /* windows */
  void		(*ReadWindowPos)(CharPtr id, Int2Ptr left, Int2Ptr top);
  void		(*WriteWindowPos)(CharPtr id, WindoW win);
  GrouP		(*InitLargeWindow)(WindoW w);
  void		(*ShowLargeWindow)(GrouP g);
  void		(*Redraw)(PaneL p);
  /* Message */
  Int2		CDECL (* Warning)(CharPtr fmt,...);
  /* Drawing */
  void		(*lMoveTo)(Int2 x, Int2 y);
  void		(*lPoint)(Int2 x, Int2 y);
  void		(*lLine)(Int2 x1, Int2 y1, Int2 x2, Int2 y2);
  void		(*lString)(Int2 x, Int2 y, CharPtr str);
#if DEB
  void		(*printRect)(CharPtr title, RectPtr rp);
  void		(*printRectL)(CharPtr title, RectPtr rp);
#endif	/* DEB */
} InterFunc, PNTR InterFuncPtr;

#ifdef KERNEL

typedef EntryPtr(void,LIB_IINIT_FUNC_TYPE,(InterFuncPtr ifp, LibHandle olh));
typedef EntryPtr(void,LIB_ITERM_FUNC_TYPE,(void));

extern Global(InterFunc) InterFuncTable;

#else /* #ifdef KERNEL */

Local(InterFuncPtr) interFuncPtr=NULL;
Local(LibHandle) ownLibHandle=NULL;     /* used in UnlockRequest */

Entry(void) LIB_IINIT_FUNC_NAME(InterFuncPtr ifp, LibHandle olh) {
  if (interFuncPtr==NULL) {
    interFuncPtr=ifp;
    ownLibHandle=olh;
  }
}

#undef screenRect
#define screenRect (*(interFuncPtr->screenRect_))

#undef dblClick
#define dblClick (*(interFuncPtr->dblClick_))

#undef sysCharWidth
#define sysCharWidth (*(interFuncPtr->sysCharWidth_))

#undef sysLineHeight
#define sysLineHeight (*(interFuncPtr->sysLineHeight_))

#undef sysDescent
#define sysDescent (*(interFuncPtr->sysDescent_))

#undef sysAscent
#define sysAscent (*(interFuncPtr->sysAscent_))

#undef vScrollBarWidth
#define vScrollBarWidth  (*(interFuncPtr->vScrollBarWidth_))

#if _UNIX
#define Nlm_hasColor (*(interFuncPtr->Nlm_hasColor_))
#endif
#if _MAC
#define Nlm_hasColorQD (*(interFuncPtr->Nlm_hasColor_))
#endif

#define ProgramFont (*(interFuncPtr->ProgramFont_))

#define SystemFont (*(interFuncPtr->SystemFont_))

#undef dialogTextHeight
#define dialogTextHeight (*(interFuncPtr->dialogTextHeight_))

#undef StringWidth
#define StringWidth interFuncPtr->StringWidth_

#undef TextWidth
#define TextWidth interFuncPtr->TextWidth_

#undef DocumentWindow
#define DocumentWindow interFuncPtr->DocumentWindow_

#undef FixedWindow
#define FixedWindow interFuncPtr->FixedWindow_

#undef SetWindowExtra
#define SetWindowExtra interFuncPtr->SetWindowExtra_

#undef GetWindowExtra
#define GetWindowExtra interFuncPtr->GetWindowExtra_

#undef ParentWindow
#define ParentWindow interFuncPtr->ParentWindow_

#undef SetAction
#define SetAction interFuncPtr->SetAction_

#undef CurrentWindow
#define CurrentWindow interFuncPtr->CurrentWindow_

#undef UseWindow
#define UseWindow interFuncPtr->UseWindow_

#undef NormalGroup
#define NormalGroup interFuncPtr->NormalGroup_

#undef HiddenGroup
#define HiddenGroup interFuncPtr->HiddenGroup_

#undef SetGroupMargins
#define SetGroupMargins interFuncPtr->SetGroupMargins_

#undef SetGroupSpacing
#define SetGroupSpacing interFuncPtr->SetGroupSpacing_

#undef Show
#define Show interFuncPtr->Show_

#undef Hide
#define Hide interFuncPtr->Hide_

#undef Enable
#define Enable interFuncPtr->Enable_

#undef Disable
#define Disable interFuncPtr->Disable_

#undef Visible
#define Visible interFuncPtr->Visible_

#undef Select
#define Select interFuncPtr->Select_

#undef SetValue
#define SetValue interFuncPtr->SetValue_

#undef GetValue
#define GetValue interFuncPtr->GetValue_

#undef SetTitle
#define SetTitle interFuncPtr->SetTitle_

#undef GetTitle
#define GetTitle interFuncPtr->GetTitle_

#undef SetStatus
#define SetStatus interFuncPtr->SetStatus_

#undef GetStatus
#define GetStatus interFuncPtr->GetStatus_

#undef SetPosition
#define SetPosition interFuncPtr->SetPosition_

#undef GetPosition
#define GetPosition interFuncPtr->GetPosition_

#undef SetOffset
#define SetOffset interFuncPtr->SetOffset_

#undef GetOffset
#define GetOffset interFuncPtr->GetOffset_

#undef Reset
#define Reset interFuncPtr->Reset_

#undef Remove
#define Remove interFuncPtr->Remove_

#undef SetNextPosition
#define SetNextPosition interFuncPtr->SetNextPosition_

#undef GetNextPosition
#define GetNextPosition interFuncPtr->GetNextPosition_

#undef Advance
#define Advance interFuncPtr->Advance_

#undef Break
#define Break interFuncPtr->Break_


#undef DialogText
#define DialogText interFuncPtr->DialogText_

#undef HiddenText
#define HiddenText interFuncPtr->HiddenText_

#undef TextLength
#define TextLength interFuncPtr->TextLength_

#undef SelectText
#define SelectText interFuncPtr->SelectText_

#undef TextSelectionRange
#define TextSelectionRange interFuncPtr->TextSelectionRange_

#undef SetTextSelect
#define SetTextSelect interFuncPtr->SetTextSelect_

#undef SingleList
#define SingleList interFuncPtr->SingleList_

#undef ListItem
#define ListItem interFuncPtr->ListItem_

#undef PopupList
#define PopupList interFuncPtr->PopupList_

#undef PopupItem
#define PopupItem interFuncPtr->PopupItem_

#undef PushButton
#define PushButton interFuncPtr->PushButton_

#undef DefaultButton
#define DefaultButton interFuncPtr->DefaultButton_

#undef CheckBox
#define CheckBox interFuncPtr->CheckBox_

#undef RadioButton
#define RadioButton interFuncPtr->RadioButton_

#undef StaticPrompt
#define StaticPrompt interFuncPtr->StaticPrompt_

#undef AlignObjects
#define AlignObjects interFuncPtr->AlignObjects_

#undef SimplePanel
#define SimplePanel interFuncPtr->SimplePanel_

#undef GetSlateVScrollBar
#define GetSlateVScrollBar interFuncPtr->GetSlateVScrollBar_

#undef SetPanelClick
#define SetPanelClick interFuncPtr->SetPanelClick_

#undef ScrollBar
#define ScrollBar interFuncPtr->ScrollBar_

#undef CorrectBarValue
#define CorrectBarValue interFuncPtr->CorrectBarValue_

#undef CorrectBarMax
#define CorrectBarMax interFuncPtr->CorrectBarMax_

#undef SetRange
#define SetRange interFuncPtr->SetRange_

#undef SetObjectExtra
#define SetObjectExtra interFuncPtr->SetObjectExtra_

#undef GetObjectExtra
#define GetObjectExtra interFuncPtr->GetObjectExtra_

#undef Parent
#define Parent interFuncPtr->Parent_

#undef Update
#define Update interFuncPtr->Update_

#undef InvalRect
#define InvalRect interFuncPtr->InvalRect_

#undef PaintRect
#define PaintRect interFuncPtr->PaintRect_

#undef ClipRect
#define ClipRect interFuncPtr->ClipRect_

#undef ResetClip
#define ResetClip interFuncPtr->ResetClip_

#undef OffsetRect
#define OffsetRect interFuncPtr->OffsetRect_

#undef LoadRect
#define LoadRect interFuncPtr->LoadRect_

#undef InsetRect
#define InsetRect interFuncPtr->InsetRect_

#undef UpsetRect
#define UpsetRect interFuncPtr->UpsetRect_

#undef UnionRect
#define UnionRect interFuncPtr->UnionRect_

#undef SectRect
#define SectRect interFuncPtr->SectRect_

#undef PtInRect
#define PtInRect interFuncPtr->PtInRect_

#undef LoadPt
#define LoadPt interFuncPtr->LoadPt_

#undef AddPt
#define AddPt interFuncPtr->AddPt_

#undef SubPt
#define SubPt interFuncPtr->SubPt_

#undef EqualPt
#define EqualPt interFuncPtr->EqualPt_

#undef FrameOval
#define FrameOval interFuncPtr->FrameOval_

#undef PaintOval
#define PaintOval interFuncPtr->PaintOval_

#undef FrameArc
#define FrameArc interFuncPtr->FrameArc_

#undef MoveTo
#define MoveTo interFuncPtr->MoveTo_

#undef LineTo
#define LineTo interFuncPtr->LineTo_

#undef SetPen
#define SetPen interFuncPtr->SetPen_

#undef DrawLine
#define DrawLine interFuncPtr->DrawLine_

#undef SelectColor
#define SelectColor interFuncPtr->SelectColor_

#undef GetColor
#define GetColor interFuncPtr->GetColor_

#undef InvertColors
#define InvertColors interFuncPtr->InvertColors_

#undef DecodeColor
#define DecodeColor interFuncPtr->DecodeColor_

#undef Black
#define Black interFuncPtr->Black_

#undef Red
#define Red interFuncPtr->Red_

#undef Green
#define Green interFuncPtr->Green_

#undef Blue
#define Blue interFuncPtr->Blue_

#undef Cyan
#define Cyan interFuncPtr->Cyan_

#undef Magenta
#define Magenta interFuncPtr->Magenta_

#undef Yellow
#define Yellow interFuncPtr->Yellow_

#undef White
#define White interFuncPtr->White_

#undef WidePen
#define WidePen interFuncPtr->WidePen_

#undef SetPenDash
#define SetPenDash interFuncPtr->SetPenDash_

#undef Solid
#define Solid interFuncPtr->Solid_

#undef Dotted
#define Dotted interFuncPtr->Dotted_

#undef Dashed
#define Dashed interFuncPtr->Dashed_

#undef ParseFont
#define ParseFont interFuncPtr->ParseFont_

#undef SelectFont
#define SelectFont interFuncPtr->SelectFont_

#undef DeleteFont
#define DeleteFont interFuncPtr->DeleteFont_

#undef Ascent
#define Ascent interFuncPtr->Ascent_

#undef Descent
#define Descent interFuncPtr->Descent_

#undef Leading
#define Leading interFuncPtr->Leading_

#undef FontHeight
#define FontHeight interFuncPtr->FontHeight_

#undef LineHeight
#define LineHeight interFuncPtr->LineHeight_

#undef PaintString
#define PaintString interFuncPtr->PaintString_

#undef DrawString
#define DrawString interFuncPtr->DrawString_

#undef PaintText
#define PaintText interFuncPtr->PaintText_

#undef RealizeWindow
#define RealizeWindow interFuncPtr->RealizeWindow_

#undef InvertMode
#define InvertMode interFuncPtr->InvertMode_

#undef CopyMode
#define CopyMode interFuncPtr->CopyMode_

#undef FrameRect
#define FrameRect interFuncPtr->FrameRect_

#undef CopyBits
#define CopyBits interFuncPtr->CopyBits_

#undef UpDownSwitch_
#undef LeftRightSwitch_
#undef SetSwitchMax_
#undef GetSwitchMax_
#undef SetSwitchParams_
#if MYSWITCH
#define SwitchCreate interFuncPtr->SwitchCreate_
#define SwitchSetValue interFuncPtr->SwitchSetValue_
#define SwitchGetValue interFuncPtr->SwitchGetValue_
#define SwitchSetMax interFuncPtr->SwitchSetMax_
#define SwitchSetParams interFuncPtr->SwitchSetParams_
#define SwitchHeight interFuncPtr->SwitchHeight_
#else
#define UpDownSwitch interFuncPtr->UpDownSwitch_
#define LeftRightSwitch interFuncPtr->LeftRightSwitch_
#define SetSwitchMax interFuncPtr->SetSwitchMax_
#define GetSwitchMax interFuncPtr->GetSwitchMax_
#define SetSwitchParams interFuncPtr->SetSwitchParams_
#endif

#undef DocumentPanel
#define DocumentPanel interFuncPtr->DocumentPanel_

#undef UpdateDocument
#define UpdateDocument interFuncPtr->UpdateDocument_

#undef AppendText
#define AppendText interFuncPtr->AppendText_

#undef ReplaceText
#define ReplaceText interFuncPtr->ReplaceText_

#undef DeleteItem
#define DeleteItem interFuncPtr->DeleteItem_

#undef AdjustDocScroll
#define AdjustDocScroll interFuncPtr->AdjustDocScroll_

#undef SetDocShade
#define SetDocShade interFuncPtr->SetDocShade_

#undef GetDocParams
#define GetDocParams interFuncPtr->GetDocParams_

#undef GetColParams
#define GetColParams interFuncPtr->GetColParams_

#undef ItemIsVisible
#define ItemIsVisible interFuncPtr->ItemIsVisible_

#undef GetScrlParams
#define GetScrlParams interFuncPtr->GetScrlParams_

#undef ProcessUpdatesFirst
#define ProcessUpdatesFirst interFuncPtr->ProcessUpdatesFirst_

#undef GetInputFileName
#define GetInputFileName interFuncPtr->GetInputFileName_

#undef GetOutputFileName
#define GetOutputFileName interFuncPtr->GetOutputFileName_

#undef ChooseColorDialog
#define ChooseColorDialog interFuncPtr->ChooseColorDialog_

/* Own functions */

#define LibraryUnlockRequest interFuncPtr->LibraryUnlockRequest

#define ReadWindowPos interFuncPtr->ReadWindowPos
#define WriteWindowPos interFuncPtr->WriteWindowPos

#define InitLargeWindow interFuncPtr->InitLargeWindow
#define ShowLargeWindow interFuncPtr->ShowLargeWindow
#define Redraw interFuncPtr->Redraw

#define Warning interFuncPtr->Warning

#define lMoveTo interFuncPtr->lMoveTo
#define lPoint interFuncPtr->lPoint
#define lLine interFuncPtr->lLine
#define lString interFuncPtr->lString

#if DEB

#define printRect interFuncPtr->printRect
#define printRectL interFuncPtr->printRectL

#endif	/* DEB */

#endif /* #ifdef KERNEL */
